@@ -33,7 +33,7 @@ ensure_bitarray(PyObject *obj)
 }
 
 /* return new bitarray of length `nbits` (with uninitialized buffer) and
-   endianness given by the PyObject 'endian' */
+   endianness given by the PyObject 'endian' (which may be Py_None) */
 static bitarrayobject *
 new_bitarray(Py_ssize_t nbits, PyObject *endian)
 {
@@ -43,7 +43,7 @@ new_bitarray(Py_ssize_t nbits, PyObject *endian)
     if ((args = PyTuple_New(2)) == NULL)
         return NULL;
 
-    /* PyTuple_SET_ITEM “steals” a reference to item */
+    /* PyTuple_SET_ITEM "steals" a reference to item */
     PyTuple_SET_ITEM(args, 0, PyLong_FromSsize_t(nbits));
     Py_INCREF(endian);
     PyTuple_SET_ITEM(args, 1, endian);
@@ -89,57 +89,51 @@ endianness, which may be 'big', 'little'.");
 /* ------------------------------- count_n ----------------------------- */
 
 /* Return the smallest index i for which a.count(vi, 0, i) == n.
-   When n exceeds the total count, return -1.  */
+   When n exceeds the total count, the result is a always negative
+   number (the negative of the total count + 1, which is useful
+   for displaying the error message). */
 static Py_ssize_t
-count_to_n(bitarrayobject *a, Py_ssize_t n, int vi)
+count_n_core(bitarrayobject *a, Py_ssize_t n, int vi)
 {
     const Py_ssize_t nbits = a->nbits;
-    unsigned char *ucbuff = (unsigned char *) a->ob_item;
-    Py_ssize_t i = 0;        /* index */
-    Py_ssize_t j = 0;        /* total count up to index */
-    Py_ssize_t block_start, block_stop, k, m;
+    uint64_t *wbuff = WBUFF(a);
+    Py_ssize_t i = 0;         /* index (result) */
+    Py_ssize_t t = 0;         /* total count up to index */
+    Py_ssize_t m;             /* popcount in each block */
 
     assert(0 <= n && n <= nbits);
     if (n == 0)
         return 0;
 
-#define BLOCK_BITS  8192
-    /* by counting big blocks we save comparisons */
+    /* by counting big blocks we save comparisons and updates */
+#define BLOCK_BITS  4096      /* block size: 4096 bits = 64 words */
     while (i + BLOCK_BITS < nbits) {
-        m = 0;
-        assert(i % 8 == 0);
-        block_start = i >> 3;
-        block_stop = block_start + (BLOCK_BITS >> 3);
-        assert(block_stop <= Py_SIZE(a));
-        for (k = block_start; k < block_stop; k++)
-            m += bitcount_lookup[ucbuff[k]];
+        m = popcnt_words(wbuff + i / 64, BLOCK_BITS / 64);
         if (!vi)
             m = BLOCK_BITS - m;
-        if (j + m >= n)
+        if (t + m >= n)
             break;
-        j += m;
+        t += m;
         i += BLOCK_BITS;
     }
 #undef BLOCK_BITS
 
-    while (i + 8 < nbits) {
-        k = i >> 3;
-        assert(k < Py_SIZE(a));
-        m = bitcount_lookup[ucbuff[k]];
+    while (i + 64 < nbits) {  /* count blocks of single (64-bit) words */
+        m = popcnt_64(wbuff[i / 64]);
         if (!vi)
-            m = 8 - m;
-        if (j + m >= n)
+            m = 64 - m;
+        if (t + m >= n)
             break;
-        j += m;
-        i += 8;
+        t += m;
+        i += 64;
     }
 
-    while (j < n && i < nbits) {
-        j += vi ? getbit(a, i) : 1 - getbit(a, i);
-        i++;
+    while (i < nbits && t < n) {
+        t += getbit(a, i++) == vi;
     }
-    if (j < n)  /* n exceeds total count */
-        return -1;
+
+    if (t < n)  /* n exceeds total count */
+        return -(t + 1);
 
     return i;
 }
@@ -158,16 +152,16 @@ count_n(PyObject *module, PyObject *args)
         PyErr_SetString(PyExc_ValueError, "non-negative integer expected");
         return NULL;
     }
-    if (n > a->nbits)  {
-        PyErr_SetString(PyExc_ValueError, "n larger than bitarray size");
-        return NULL;
-    }
-    i = count_to_n(a, n, vi);        /* do actual work here */
+    if (n > a->nbits)
+        return PyErr_Format(PyExc_ValueError, "n = %zd larger than bitarray "
+                            "size (len(a) = %zd)", n, a->nbits);
+
+    i = count_n_core(a, n, vi);        /* do actual work here */
+
+    if (i < 0)
+        return PyErr_Format(PyExc_ValueError, "n = %zd exceeds total count "
+                            "(a.count(%d) = %zd)", n, vi, -i - 1);
 
-    if (i < 0) {
-        PyErr_SetString(PyExc_ValueError, "n exceeds total count");
-        return NULL;
-    }
     return PyLong_FromSsize_t(i);
 }
 
@@ -175,7 +169,7 @@ PyDoc_STRVAR(count_n_doc,
 "count_n(a, n, value=1, /) -> int\n\
 \n\
 Return lowest index `i` for which `a[:i].count(value) == n`.\n\
-Raises `ValueError`, when n exceeds total count (`a.count(value)`).");
+Raises `ValueError` when `n` exceeds total count (`a.count(value)`).");
 
 /* ----------------------------- right index --------------------------- */
 
@@ -193,26 +187,27 @@ find_last(bitarrayobject *self, int vi, Py_ssize_t a, Py_ssize_t b)
         return -1;
 
     /* the logic here is the same as in find_bit() in _bitarray.c */
-#ifdef PY_UINT64_T
     if (n > 64) {
-        Py_ssize_t word_a = (a + 63) / 64;
-        Py_ssize_t word_b = b / 64;
-        PY_UINT64_T *wbuff = (PY_UINT64_T *) self->ob_item, w = vi ? 0 : ~0;
+        const Py_ssize_t wa = (a + 63) / 64;  /* word range(wa, wb) */
+        const Py_ssize_t wb = b / 64;
+        const uint64_t *wbuff = WBUFF(self);
+        const uint64_t w = vi ? 0 : ~0;
 
-        if ((res = find_last(self, vi, 64 * word_b, b)) >= 0)
+        if ((res = find_last(self, vi, 64 * wb, b)) >= 0)
             return res;
 
-        for (i = word_b - 1; i >= word_a; i--) {  /* skip uint64 words */
+        for (i = wb - 1; i >= wa; i--) {  /* skip uint64 words */
             if (w ^ wbuff[i])
                 return find_last(self, vi, 64 * i, 64 * i + 64);
         }
-        return find_last(self, vi, a, 64 * word_a);
+        return find_last(self, vi, a, 64 * wa);
     }
-#endif
+
     if (n > 8) {
-        Py_ssize_t byte_a = BYTES(a);
-        Py_ssize_t byte_b = b / 8;
-        char *buff = self->ob_item, c = vi ? 0 : ~0;
+        const Py_ssize_t byte_a = BYTES(a);  /* byte range(byte_a, byte_b) */
+        const Py_ssize_t byte_b = b / 8;
+        const char *buff = self->ob_item;
+        const char c = vi ? 0 : ~0;
 
         if ((res = find_last(self, vi, 8 * byte_b, b)) >= 0)
             return res;
@@ -224,7 +219,7 @@ find_last(bitarrayobject *self, int vi, Py_ssize_t a, Py_ssize_t b)
         }
         return find_last(self, vi, a, 8 * byte_a);
     }
-    assert(n <= 8);
+
     for (i = b - 1; i >= a; i--) {
         if (getbit(self, i) == vi)
             return i;
@@ -262,19 +257,20 @@ static PyObject *
 parity(PyObject *module, PyObject *obj)
 {
     bitarrayobject *a;
-    unsigned char par = 0;
+    uint64_t par = 0;
     Py_ssize_t i;
 
     if (ensure_bitarray(obj) < 0)
         return NULL;
 
     a = (bitarrayobject *) obj;
-    for (i = 0; i < a->nbits / 8; i++)
-        par ^= a->ob_item[i];
-    if (a->nbits % 8)
-        par ^= zeroed_last_byte(a);
-
-    return PyLong_FromLong((long) bitcount_lookup[par] % 2);
+    for (i = 0; i < a->nbits / 64; i++)
+        par ^= WBUFF(a)[i];
+    if (a->nbits % 64)
+        par ^= zlw(a);
+    for (i = 32; i > 0; i /= 2)
+        par ^= par >> i;
+    return PyLong_FromLong(par & 1);
 }
 
 PyDoc_STRVAR(parity_doc,
@@ -304,10 +300,10 @@ same_size_endian(bitarrayobject *a, bitarrayobject *b)
 static PyObject *
 binary_function(PyObject *args, const char *format, const char oper)
 {
-    Py_ssize_t cnt = 0, s, i;
+    Py_ssize_t cnt = 0, cwords, i;
     bitarrayobject *a, *b;
-    unsigned char *buff_a, *buff_b;
-    int r;
+    uint64_t *wbuff_a, *wbuff_b;
+    int rbits;
 
     if (!PyArg_ParseTuple(args, format,
                           bitarray_type_obj, (PyObject *) &a,
@@ -316,51 +312,49 @@ binary_function(PyObject *args, const char *format, const char oper)
     if (same_size_endian(a, b) < 0)
         return NULL;
 
-    buff_a = (unsigned char *) a->ob_item;
-    buff_b = (unsigned char *) b->ob_item;
-    s = a->nbits / 8;       /* number of whole bytes in buffer */
-    r = a->nbits % 8;       /* remaining bits  */
+    wbuff_a = WBUFF(a);
+    wbuff_b = WBUFF(b);
+    cwords = a->nbits / 64;     /* number of complete 64-bit words */
+    rbits = a->nbits % 64;      /* remaining bits  */
 
     switch (oper) {
-#define UZ(x)  ((unsigned char) zeroed_last_byte(x))
     case '&':                   /* count and */
-        for (i = 0; i < s; i++)
-            cnt += bitcount_lookup[buff_a[i] & buff_b[i]];
-        if (r)
-            cnt += bitcount_lookup[UZ(a) & UZ(b)];
+        for (i = 0; i < cwords; i++)
+            cnt += popcnt_64(wbuff_a[i] & wbuff_b[i]);
+        if (rbits)
+            cnt += popcnt_64(zlw(a) & zlw(b));
         break;
 
     case '|':                   /* count or */
-        for (i = 0; i < s; i++)
-            cnt += bitcount_lookup[buff_a[i] | buff_b[i]];
-        if (r)
-            cnt += bitcount_lookup[UZ(a) | UZ(b)];
+        for (i = 0; i < cwords; i++)
+            cnt += popcnt_64(wbuff_a[i] | wbuff_b[i]);
+        if (rbits)
+            cnt += popcnt_64(zlw(a) | zlw(b));
         break;
 
     case '^':                   /* count xor */
-        for (i = 0; i < s; i++)
-            cnt += bitcount_lookup[buff_a[i] ^ buff_b[i]];
-        if (r)
-            cnt += bitcount_lookup[UZ(a) ^ UZ(b)];
+        for (i = 0; i < cwords; i++)
+            cnt += popcnt_64(wbuff_a[i] ^ wbuff_b[i]);
+        if (rbits)
+            cnt += popcnt_64(zlw(a) ^ zlw(b));
         break;
 
     case 'a':                   /* any and */
-        for (i = 0; i < s; i++) {
-            if (buff_a[i] & buff_b[i])
+        for (i = 0; i < cwords; i++) {
+            if (wbuff_a[i] & wbuff_b[i])
                 Py_RETURN_TRUE;
         }
-        return PyBool_FromLong(r && (UZ(a) & UZ(b)));
+        return PyBool_FromLong(rbits && (zlw(a) & zlw(b)));
 
     case 's':                   /* is subset */
-        for (i = 0; i < s; i++) {
-            if ((buff_a[i] & buff_b[i]) != buff_a[i])
+        for (i = 0; i < cwords; i++) {
+            if ((wbuff_a[i] & wbuff_b[i]) != wbuff_a[i])
                 Py_RETURN_FALSE;
         }
-        return PyBool_FromLong(r == 0 || (UZ(a) & UZ(b)) == UZ(a));
+        return PyBool_FromLong(rbits == 0 || (zlw(a) & zlw(b)) == zlw(a));
 
     default:
         Py_UNREACHABLE();
-#undef UZ
     }
     return PyLong_FromSsize_t(cnt);
 }
@@ -412,9 +406,10 @@ iteration is stopped as soon as one mismatch is found.");
 static PyObject *
 correspond_all(PyObject *module, PyObject *args)
 {
-    Py_ssize_t nff = 0, nft = 0, ntf = 0, ntt = 0, i, s;
+    Py_ssize_t nff = 0, nft = 0, ntf = 0, ntt = 0, cwords, i;
     bitarrayobject *a, *b;
-    unsigned char u, v, not_u, not_v;
+    uint64_t u, v, not_u, not_v;
+    int rbits;
 
     if (!PyArg_ParseTuple(args, "O!O!:_correspond_all",
                           bitarray_type_obj, (PyObject *) &a,
@@ -422,28 +417,31 @@ correspond_all(PyObject *module, PyObject *args)
         return NULL;
     if (same_size_endian(a, b) < 0)
         return NULL;
-    s = a->nbits / 8;       /* number of whole bytes in buffer */
 
-    for (i = 0; i < s; i++) {
-        u = a->ob_item[i];
-        v = b->ob_item[i];
+    cwords = a->nbits / 64;     /* complete 64-bit words */
+    rbits = a->nbits % 64;      /* remaining bits */
+
+    for (i = 0; i < cwords; i++) {
+        u = WBUFF(a)[i];
+        v = WBUFF(b)[i];
         not_u = ~u;
         not_v = ~v;
-        nff += bitcount_lookup[not_u & not_v];
-        nft += bitcount_lookup[not_u & v];
-        ntf += bitcount_lookup[u & not_v];
-        ntt += bitcount_lookup[u & v];
+        nff += popcnt_64(not_u & not_v);
+        nft += popcnt_64(not_u & v);
+        ntf += popcnt_64(u & not_v);
+        ntt += popcnt_64(u & v);
     }
-    if (a->nbits % 8) {
-        unsigned char mask = ones_table[IS_BE(a)][a->nbits % 8];
-        u = a->ob_item[s];
-        v = b->ob_item[s];
+
+    if (rbits) {
+        u = zlw(a);
+        v = zlw(b);
         not_u = ~u;
         not_v = ~v;
-        nff += bitcount_lookup[not_u & not_v & mask];
-        nft += bitcount_lookup[not_u & v & mask];
-        ntf += bitcount_lookup[u & not_v & mask];
-        ntt += bitcount_lookup[u & v & mask];
+        /* for nff we need to substract the number of unused 1 bits */
+        nff += popcnt_64(not_u & not_v) - (64 - rbits);
+        nft += popcnt_64(not_u & v);
+        ntf += popcnt_64(u & not_v);
+        ntt += popcnt_64(u & v);
     }
     return Py_BuildValue("nnnn", nff, nft, ntf, ntt);
 }
@@ -962,7 +960,7 @@ next_char(PyObject *iter)
         return -1;
     }
 
-#ifdef IS_PY3K
+#if IS_PY3K
     if (PyLong_Check(item))
         c = (unsigned char) PyLong_AsLong(item);
 #else
@@ -1030,18 +1028,19 @@ byte_length(Py_ssize_t i)
     return n;
 }
 
-/* starting from byte i count the remaining population in bitarray buffer */
+/* Starting from word index `i`, count the remaining population in bitarray
+   buffer.  Equivalent to:  a[64 * i:].count()  */
 static Py_ssize_t
-count_from(bitarrayobject *a, Py_ssize_t i)
+count_from_word(bitarrayobject *a, Py_ssize_t i)
 {
+    const Py_ssize_t nbits = a->nbits;
     Py_ssize_t cnt = 0;
 
-    while (i < a->nbits / 8)
-        cnt += bitcount_lookup[(unsigned char) a->ob_item[i++]];
-
-    if (a->nbits % 8)
-        cnt += bitcount_lookup[(unsigned char) zeroed_last_byte(a)];
-
+    if (64 * i >= nbits)
+        return 0;
+    cnt += popcnt_words(WBUFF(a) + i, nbits / 64 - i);
+    if (nbits % 64)
+        cnt += popcnt_64(zlw(a));
     return cnt;
 }
 
@@ -1056,10 +1055,12 @@ count_from(bitarrayobject *a, Py_ssize_t i)
 #define BSI(n)  (((Py_ssize_t) 1) << (8 * (n) - 3))
 
 /* segment size in bytes - Although of little practical value, the code
-   below will also work when changing SEGSIZE to 1, 2, 4, 8 or 16, as long
-   as a multiple of SEGSIZE is 32.  The size 32 is rooted in the fact that
-   a bitarray of 32 bytes (256 bits) can be indexed with one index byte
-   (BSI(1) = 32).  Our entire 'sc' format is constructed around this. */
+   below will also work for SEGSIZE values of: 8, 16 and 32
+   BSI(1) = 32 must be divisible by SEGSIZE.
+   SEGSIZE must also be a multiple of the word size (sizeof(uint64_t) = 8).
+   The size 32 is rooted in the fact that a bitarray of 32 bytes (256 bits)
+   can be indexed with one index byte (BSI(1) = 32).  Our entire 'sc' format
+   is constructed around this. */
 #define SEGSIZE  32
 
 /* number of 256 bit segments given nbits */
@@ -1108,42 +1109,70 @@ count_from(bitarrayobject *a, Py_ssize_t i)
 static Py_ssize_t *
 sc_calc_rts(bitarrayobject *a)
 {
-    const Py_ssize_t n_seg = NSEG(a->nbits);  /* number of segments */
-    const Py_ssize_t c_seg = a->nbits / (8 * SEGSIZE);  /* complete segs */
+    const Py_ssize_t nbits = a->nbits;
+    const Py_ssize_t n_seg = NSEG(nbits);     /* number of segments */
+    const Py_ssize_t c_seg = nbits / (8 * SEGSIZE);  /* complete segments */
     char zeros[SEGSIZE];                      /* segment with only zeros */
     Py_ssize_t cnt = 0;                       /* current count */
-    char *buff;
-    Py_ssize_t *res, i, j;
+    char *buff;                               /* buffer in current segment */
+    Py_ssize_t *res, m;
 
-    assert(n_seg == c_seg || n_seg == c_seg + 1);
     memset(zeros, 0x00, SEGSIZE);
     res = (Py_ssize_t *) PyMem_Malloc((size_t)
                                       sizeof(Py_ssize_t) * (n_seg + 1));
     if (res == NULL)
         return (Py_ssize_t *) PyErr_NoMemory();
 
-    /* all complete segments */
-    for (j = 0, buff = a->ob_item; j < c_seg; j++, buff += SEGSIZE) {
-        res[j] = cnt;
-        assert(buff - a->ob_item + SEGSIZE <= Py_SIZE(a));
-        if (memcmp(buff, zeros, SEGSIZE)) { /* segment has not only zeros */
-            for (i = 0; i < SEGSIZE; i++)
-                cnt += bitcount_lookup[(unsigned char) buff[i]];
-        }
+    for (m = 0; m < c_seg; m++) {  /* loop all complete segments */
+        res[m] = cnt;
+        buff = a->ob_item + m * SEGSIZE;
+        assert((m + 1) * SEGSIZE <= Py_SIZE(a));
+
+        if (memcmp(buff, zeros, SEGSIZE))  /* segment has not only zeros */
+            cnt += popcnt_words((uint64_t *) buff, SEGSIZE / 8);
     }
-    assert(buff - a->ob_item == SEGSIZE * c_seg);
     res[c_seg] = cnt;
 
     if (n_seg > c_seg) {           /* we have a final partial segment */
-        assert(Py_SIZE(a) <= SEGSIZE * n_seg);
-        assert(a->nbits && a->nbits < 8 * SEGSIZE * n_seg);
-
-        cnt += count_from(a, SEGSIZE * c_seg);
+        cnt += count_from_word(a, (SEGSIZE / 8) * c_seg);
         res[n_seg] = cnt;
     }
     return res;
 }
 
+/* expose sc_calc_rts() to Python during debug mode for testing */
+#ifndef NDEBUG
+static PyObject *
+sc_rts(PyObject *module, PyObject *obj)
+{
+    PyObject *list;
+    bitarrayobject *a;
+    Py_ssize_t *rts, i;
+
+    if (ensure_bitarray(obj) < 0)
+        return NULL;
+
+    a = (bitarrayobject *) obj;
+    if ((rts = sc_calc_rts(a)) == NULL)
+        return NULL;
+
+    if ((list = PyList_New(NSEG(a->nbits) + 1)) == NULL)
+        return NULL;
+
+    for (i = 0; i <= NSEG(a->nbits); i++) {
+        PyObject *item = PyLong_FromSsize_t(rts[i]);
+        if (item == NULL) {
+            Py_DECREF(list);
+            return NULL;
+        }
+        PyList_SET_ITEM(list, i, item);
+    }
+    PyMem_Free(rts);
+    return list;
+}
+#endif  /* NDEBUG */
+
+
 /* Equivalent to the Python expression:
 
       a.count(1, 8 * offset, 8 * offset + (1 << (8 * n)))
@@ -1263,9 +1292,6 @@ sc_write_indices(char *str, bitarrayobject *a, Py_ssize_t *rts,
     Py_UNREACHABLE();
 }
 
-#undef SEGSIZE
-#undef NSEG
-
 /* Write one sparse block (from `offset`, and up to `k` one bits).
    Return number of bytes written to buffer `str` (encoded block size). */
 static Py_ssize_t
@@ -1904,7 +1930,7 @@ chdi_traverse(chdi_obj *it, visitproc visit, void *arg)
 }
 
 static PyTypeObject CHDI_Type = {
-#ifdef IS_PY3K
+#if IS_PY3K
     PyVarObject_HEAD_INIT(NULL, 0)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -1973,19 +1999,23 @@ static PyMethodDef module_functions[] = {
                                            METH_VARARGS, vl_decode_doc},
     {"canonical_decode",
                   (PyCFunction) chdi_new,  METH_VARARGS, chdi_doc},
+#ifndef NDEBUG
+    /* functionality exposed in debug mode for testing */
+    {"_sc_rts",   (PyCFunction) sc_rts,    METH_O,       0},
+#endif
     {NULL,        NULL}  /* sentinel */
 };
 
 /******************************* Install Module ***************************/
 
-#ifdef IS_PY3K
+#if IS_PY3K
 static PyModuleDef moduledef = {
     PyModuleDef_HEAD_INIT, "_util", 0, -1, module_functions,
 };
 #endif
 
 PyMODINIT_FUNC
-#ifdef IS_PY3K
+#if IS_PY3K
 PyInit__util(void)
 #else
 init_util(void)
@@ -2000,7 +2030,7 @@ init_util(void)
     if (bitarray_type_obj == NULL)
         goto error;
 
-#ifdef IS_PY3K
+#if IS_PY3K
     m = PyModule_Create(&moduledef);
 #else
     m = Py_InitModule3("_util", module_functions, 0);
@@ -2012,7 +2042,11 @@ init_util(void)
         goto error;
     Py_SET_TYPE(&CHDI_Type, &PyType_Type);
 
-#ifdef IS_PY3K
+#ifndef NDEBUG  /* expose segment size for testing */
+    PyModule_AddObject(m, "_SEGSIZE", PyLong_FromSsize_t(SEGSIZE));
+#endif
+
+#if IS_PY3K
     return m;
  error:
     return NULL;
